@@ -1,7 +1,7 @@
 #!/usr/bin/perl
 
 package Directory::Transactional::TXN::Root;
-our $VERSION = '0.03';
+our $VERSION = '0.04';
 
 use Squirrel;
 
@@ -31,9 +31,12 @@ has backup => (
 
 sub _build_backup {
 	my $self = shift;
-	my $dir = File::Spec->catdir( $self->manager->_backups, $self->id );
-	make_path($dir);
-	return $dir;
+	File::Spec->catdir( $self->manager->_backups, $self->id );
+}
+
+sub create_backup_dir {
+	my $self = shift;
+	make_path($self->backup);
 }
 
 sub find_lock {
